/************************************************************************/
/*                                                                      */
/*      Tile Pages WE - Generic WebExtension - Background Page          */
/*                                                                      */
/*      Javascript for Background Page                                  */
/*                                                                      */
/*      Last Edit - 10 Jan 2019                                         */
/*                                                                      */
/*      Copyright (C) 2016-2019 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Refer to Google Chrome developer documentation:                     */
/*                                                                      */
/*  https://developer.chrome.com/extensions/overview                    */
/*  https://developer.chrome.com/extensions/content_scripts             */
/*  https://developer.chrome.com/extensions/messaging                   */
/*  https://developer.chrome.com/extensions/optionsV2                   */
/*  https://developer.chrome.com/extensions/contentSecurityPolicy       */
/*                                                                      */
/*  https://developer.chrome.com/extensions/manifest                    */
/*  https://developer.chrome.com/extensions/declare_permissions         */
/*  https://developer.chrome.com/extensions/match_patterns              */
/*                                                                      */
/*  https://developer.chrome.com/extensions/browserAction               */
/*  https://developer.chrome.com/extensions/commands                    */
/*  https://developer.chrome.com/extensions/contextMenus                */
/*  https://developer.chrome.com/extensions/extension                   */
/*  https://developer.chrome.com/extensions/notifications               */
/*  https://developer.chrome.com/extensions/runtime                     */
/*  https://developer.chrome.com/extensions/storage                     */
/*  https://developer.chrome.com/extensions/tabs                        */
/*  https://developer.chrome.com/extensions/windows                     */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Notes on Tile Pages WE Operation                                    */
/*                                                                      */
/*  1. The basic idea of Tile Pages WE is to provide 'split browser'    */
/*     functionality, to allow viewing of multiple pages in a tiled     */
/*     layout within the area of the browser window.                    */
/*                                                                      */
/*  2. Each layout is created inside a separate tab within which the    */
/*     individual tiles are dynamically created as <iframe> elements    */
/*     within the container page (moz-extension://.../container.html).  */
/*                                                                      */
/*  3. Using normal <iframe> elements presents several issues:          */
/*                                                                      */
/*     - An <iframe> must have a sandbox attribute to guard against     */
/*       security exploits. This limits what a framed web page can do,  */
/*       but does not prevent execution of add-on content scripts.      */
/*                                                                      */
/*     - An <iframe> sandbox attribute must never have both             */
/*       'allow-same-orign' and 'allow-scripts' at the same time,       */
/*       since this allows the framed web page to remove the sandbox.   */
/*                                                                      */
/*     - Tile Pages WE sets the sandbox attribute of a tiled <iframe>   */
/*       in the container page to 'allow-forms allow-same-origin'.      */
/*       The 'allow-forms' value is required to allow submission of     */
/*       user input (e.g. searches). The 'allow-same-origin' value is   */
/*       required because of a bug in Firefox that otherwise prevents   */
/*       execution of content scripts, but may be removed in future.    */
/*                                                                      */
/*     - An <iframe> sandbox attribute will not allow a web page to     */
/*       break out of the <iframe>. Consequently, some web pages use    */
/*       an 'X-Frame-Options' header or a CSP header 'frame-ancestors'  */
/*       directive to avoid being loaded into an <iframe>.              */
/*                                                                      */
/*     - Tile Pages WE removes the 'X-Frame-Options' headers and CSP    */
/*       header 'frame-ancestors' directives from web pages that are    */
/*       loaded into a tiled <iframe> in the container page, but not    */
/*       from web pages that are loaded into a nested <iframe> inside   */
/*       a web page. Security should be maintained since:               */
/*                                                                      */
/*       > the container page is an add-on page and is not hostile.     */
/*       > the container page (moz-extension:) has a different origin   */
/*         from those of the framed web pages (http: or https:).        */
/*       > the sandbox prevents scripts running in framed web pages.    */
/*       > the sandbox applies to all nested frames.                    */
/*                                                                      */
/*  5. In the longer term, Tile Pages WE may evolve to use a special    */
/*     <iframe mozbrowser remote> element that is only available in     */
/*     Firefox. This special <iframe> acts like a top-level browser     */
/*     frame and runs in a separate process, which avoids the issues    */
/*     with normal <iframe> elements.                                   */
/*                                                                      */
/*  6. Tile Pages WE tries to load a content script in each web page    */
/*     that is loaded in an <iframe> element, but this is not possible  */
/*     for a few special web pages (e.g. 'about:blank').                */
/*                                                                      */
/*  7. Tile Page WE handles content script no-loads as follows:         */
/*                                                                      */
/*     - When a web page starts loading, the styling of the <iframe>    */
/*       element is set to 'pointer-events: none'.                      */
/*                                                                      */
/*     - When a web page completes loading, if the content script was   */
/*       loaded, it sends a message which causes the styling of the     */
/*       <iframe> element to be reset to 'pointer-events: auto'.        */
/*                                                                      */
/*     - If the content script was not loaded, then any mouse events    */
/*       (e.g. mousedown, dblclick, dragover, drop and mouseleave)      */
/*       intended for the <iframe> element will be handled by the       */
/*       tile container <div> element.                                  */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Layout String Formats                                               */
/*                                                                      */
/*  Layout  Debut  Record                                               */
/*   Type    Ver   Format                                               */
/*                                                                      */
/*    all    1.0   layouttype,panelcount;                               */
/*                                                                      */
/*    all    1.0   hsplit,ratio,childcount;                             */
/*    all    1.0   vsplit,ratio,childcount;                             */
/*                                                                      */
/*     0     1.0   panel,ratio;                                         */
/*     1     1.0   panel,ratio,{pageurl};                               */
/*     2     1.0   panel,ratio,{pageurl},scrollX,scrollY;               */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var enableInspect = false;

var isFirefox;
var ffVersion;

var platformOS;

var topMargin;
var lrbMargin;

var blankPage;

var showSubmenu;
var assignTabs,assignLeft;
var startAction;
var homePage;
var xFrameOptions,cspFrameAncestors;

var layoutTabId = -1;
var layoutTabReady = false;  /* updated from container.js */

var panelCount = 0;  /* updated from container.js */

var tiledPageUrls = new Array();  /* updated from container.js */
var tiledPageTitles = new Array();  /* updated from container.js */

var viewTiled = true;  /* updated from container.js */
var syncScroll = false;  /* updated from container.js */

var dialogWindowId = -1;
var dialogParentWindowId = -1;

var dialogParentWindowLeft;
var dialogParentWindowTop;
var dialogParentWindowHeight;
var dialogParentWindowWidth;

var debugEnable = true;

/************************************************************************/

/* Initialize on browser startup */

chrome.runtime.getPlatformInfo(
function(PlatformInfo)
{
    platformOS = PlatformInfo.os;
    
    chrome.storage.local.set({ "environment-platformos": platformOS });
    
    isFirefox = (navigator.userAgent.indexOf("Firefox") >= 0);

    chrome.storage.local.set({ "environment-isfirefox": isFirefox });
    
    if (isFirefox)
    {
        chrome.runtime.getBrowserInfo(
        function(info)
        {
            ffVersion = info.version.substr(0,info.version.indexOf("."));
            
            chrome.storage.local.set({ "environment-ffversion": ffVersion });
            
            initialize();
        });
    }
    else initialize();
});

function initialize()
{
    if (platformOS == "win") { topMargin = 0; lrbMargin = 7; }
    else if (platformOS == "linux") { topMargin = 0; lrbMargin = 0; }
    else if (platformOS == "mac") { topMargin = 0; lrbMargin = 0; }
    else { topMargin = 0; lrbMargin = 0; }

    chrome.storage.local.get(null,
    function(object)
    {
        var defaultDesign,defaultName,layoutString;
        
        /* Initialize or migrate options */
        
        /* General options */

        if (!("options-shownavbar" in object)) object["options-shownavbar"] = true;
        if (!("options-settitle" in object)) object["options-settitle"] = false;
        if (!("options-showsubmenu" in object)) object["options-showsubmenu"] = true;
        
        if (!("options-assigntabs" in object)) object["options-assigntabs"] = true;
        if (!("options-assignleft" in object)) object["options-assignleft"] = false;
        
        if (!("options-startaction" in object)) object["options-startaction"] = 0;
        
		/* Page Settings options */
        
        if (!("options-homepage" in object)) object["options-homepage"] = "";
        
        if (!("options-openlinks" in object)) object["options-openlinks"] = 3;
        
        if (!("options-syncactive" in object)) object["options-syncactive"] = false;
        
		/* Advanced options */
        
        if (!("options-xframeoptions" in object)) object["options-xframeoptions"] = false;
        if (!("options-cspframeancestors" in object)) object["options-cspframeancestors"] = false;
        
        /* Layout options */
        
        if (!("layout-defaultdesign" in object)) object["layout-defaultdesign"] = "2vert";
        if (!("layout-defaultname" in object)) object["layout-defaultname"] = "";
        
        if (!("layout-savepages" in object)) object["layout-savepages"] = false;
        if (!("layout-savescroll" in object)) object["layout-savescroll"] = false;
        
        if (!("urlbar-suggestions" in object)) object["urlbar-suggestions"] = new Array();
        
        /* Update stored options */
        
        chrome.storage.local.set(object);
        
        /* Initialize local options */
        
        showSubmenu = object["options-showsubmenu"];
        
        assignTabs = object["options-assigntabs"];
        assignLeft = object["options-assignleft"];
        
        startAction = object["options-startaction"];
        
        homePage = object["options-homepage"];
        
        xFrameOptions = object["options-xframeoptions"];
        cspFrameAncestors = object["options-cspframeancestors"];
        
        homePage = homePage.replace(/\/$/,"");
        
        /* Add context menu items */
        
        chrome.tabs.query({ lastFocusedWindow: true, active: true },
        function(tabs)
        {
            createContextMenu(tabs[0].id);
        });
        
        /* Add listeners */
        
        addListeners();
        
        /* Open restore or default layout */
        
        if (startAction == 0)
        {
            chrome.storage.local.remove("layout-restore");
            
            chrome.tabs.query({ windowType: "normal" },
            function(tabs)
            {
                var i;
                
                /* Remove empty container page (if any) after Firefox session restore */
                
                for (i = 0; i < tabs.length; i++)
                    if (tabs[i].url == chrome.runtime.getURL("container.html")) chrome.tabs.remove(tabs[i].id); 
            });
        }
        else if (startAction == 1)
        {
            defaultDesign = object["layout-defaultdesign"];
            defaultName = object["layout-defaultname"];
            
            if (defaultName != "") layoutString = object["layout-string-" + defaultName];
            
            if (defaultDesign == "saved" && defaultName != "" && layoutString != null) openLayout(defaultName,layoutString);
            else newLayout(defaultDesign);
        }
        else if (startAction == 2 && "layout-restore" in object)
        {
            openLayout("",object["layout-restore"]);
        }
    });
}

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Storage changed listener */
    
    chrome.storage.onChanged.addListener(
    function(changes,areaName)
    {
        chrome.storage.local.get(null,
        function(object)
        {
            showSubmenu = object["options-showsubmenu"];
            
            assignTabs = object["options-assigntabs"];
            assignLeft = object["options-assignleft"];
            
            startAction = object["options-startaction"];
            
            homePage = object["options-homepage"];
            
            xFrameOptions = object["options-xframeoptions"];
            cspFrameAncestors = object["options-cspframeancestors"];
            
            homePage = homePage.replace(/\/$/,"");
            
            if ("options-showsubmenu" in changes)
            {
                chrome.tabs.query({ lastFocusedWindow: true, active: true },
                function(tabs)
                {
                    createContextMenu(tabs[0].id);
                });
            }
        });
    });
    
    /* Browser action listener */
    
    chrome.browserAction.onClicked.addListener(
    function(tab)
    {
        chrome.windows.getLastFocused({ populate: true },
        function(focuswin)
        {
            if (layoutTabId == -1)  /* no layout */
            {
                chrome.storage.local.get(null,
                function(object)
                {
                    var defaultDesign,defaultName,layoutString;
                    
                    defaultDesign = object["layout-defaultdesign"];
                    defaultName = object["layout-defaultname"];
                    
                    if (defaultName != "") layoutString = object["layout-string-" + defaultName];
                    
                    if (defaultDesign == "saved" && defaultName != "" && layoutString != null) openLayout(defaultName,layoutString);
                    else newLayout(defaultDesign);
                });
            }
            else if (tab.id == layoutTabId)  /* layout tab */
            {
                toggleView(tab.id);
            }
            else  /* other tab */
            {
                chrome.tabs.update(layoutTabId,{ active: true });
            }
        });
    });
    
    /* Keyboard command listener */
    
    chrome.commands.onCommand.addListener(
    function(command)
    {
        chrome.tabs.query({ lastFocusedWindow: true, active: true },
        function(tabs)
        {
            switch (command)
            {
                case "newcloselayout":
                
                    if (layoutTabId == -1)  /* no layout */
                    {
                        chrome.storage.local.get(null,
                        function(object)
                        {
                            var defaultDesign,defaultName,layoutString;
                            
                            defaultDesign = object["layout-defaultdesign"];
                            defaultName = object["layout-defaultname"];
                            
                            if (defaultName != "") layoutString = object["layout-string-" + defaultName];
                            
                            if (defaultDesign == "saved" && defaultName != "" && layoutString != null) openLayout(defaultName,layoutString);
                            else newLayout(defaultDesign);
                        });
                    }
                    else if (tabs[0].id == layoutTabId)  /* layout tab */
                    {
                        closeLayout();
                    }
                    else  /* other tab */
                    {
                        chrome.tabs.update(layoutTabId,{ active: true });
                    }
                    
                    break;
                    
                case "untilepage":
                    
                    if (layoutTabId != -1)
                    {
                        chrome.tabs.sendMessage(layoutTabId, { type: "untilePage" },{ frameId: 0 });
                    }
                    
                    break;
                    
                case "syncscroll":
                    
                    if (layoutTabId != -1)
                    {
                        toggleSync(layoutTabId);
                    }
                    
                    break;
                    
                case "viewtiled":
                    
                    if (layoutTabId != -1)
                    {
                        toggleView(layoutTabId);
                    }
                    
                    break;
            }
        });
    });
    
    /* Context menu listener */
    
    chrome.contextMenus.onClicked.addListener(
    function(info,tab)
    {
        var direction,url;
        var matches = new Array();
        
        if (info.menuItemId.substr(0,8) == "tilepage") matches = info.menuItemId.match(/(tilepage-[^-]+-[^-]+-?)(\d+)?/);
        else if (info.menuItemId.substr(0,10) == "assignpage") matches = info.menuItemId.match(/(assignpage-[^-]+-?)(\d+)?/);
        else matches[1] = info.menuItemId;
        
        switch (matches[1])
        {
            /* Page or button context menu */
            
            case "newcloselayout":
                
                closeLayout();
                
                break;
                
            case "newlayout-default":
                
                chrome.storage.local.get(null,
                function(object)
                {
                    var defaultDesign,defaultName,layoutString;
                    
                    defaultDesign = object["layout-defaultdesign"];
                    defaultName = object["layout-defaultname"];
                    
                    if (defaultName != "") layoutString = object["layout-string-" + defaultName];
                    
                    if (defaultDesign == "saved" && defaultName != "" && layoutString != null) openLayout(defaultName,layoutString);
                    else newLayout(defaultDesign);
                });
                
                break;
                
            case "newlayout-2vert":
            case "newlayout-2horiz":
            case "newlayout-3vert":
            case "newlayout-3horiz":
            case "newlayout-4vertgrid":
            case "newlayout-4horizgrid":
            case "newlayout-6vertgrid":
            case "newlayout-6horizgrid":
            case "newlayout-allvert":
            case "newlayout-allhoriz":
            case "newlayout-allvertgrid":
            case "newlayout-allhorizgrid":
                
                newLayout(info.menuItemId.substr(10));
                
                break;
                
            case "opensavelayout":
                
                if (layoutTabId == -1) openLayoutDialog();
                else saveLayoutDialog();
                
                break;
                
            case "deletelayout":
                
                deleteLayoutDialog();
                
                break;
                
            case "defaultlayout-saved":
            
                defaultLayoutDialog();
                
                chrome.storage.local.get(null,
                function(object)
                {
                    chrome.contextMenus.update("defaultlayout-saved",{ checked: (object["layout-defaultdesign"] == "saved") });
                });
                
                break;
                
            case "defaultlayout-2vert":
            case "defaultlayout-2horiz":
            case "defaultlayout-3vert":
            case "defaultlayout-3horiz":
            case "defaultlayout-4vertgrid":
            case "defaultlayout-4horizgrid":
            case "defaultlayout-6vertgrid":
            case "defaultlayout-6horizgrid":
            case "defaultlayout-allvert":
            case "defaultlayout-allhoriz":
            case "defaultlayout-allvertgrid":
            case "defaultlayout-allhorizgrid":
            
                clearDefaultLayoutCheckMarks();
                chrome.contextMenus.update(info.menuItemId,{ checked: true });
                chrome.storage.local.set({ "layout-defaultdesign": info.menuItemId.substr(14), "layout-defaultname": "" });
                
                break;
                
            case "tilepage-above-link":
            case "tilepage-left-link":
            case "tilepage-right-link":
            case "tilepage-below-link":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                chrome.tabs.sendMessage(layoutTabId, { type: "tilePage", direction: direction, url: info.linkUrl },{ frameId: 0 });
                
                break;
                
            case "tilepage-above-home":
            case "tilepage-left-home":
            case "tilepage-right-home":
            case "tilepage-below-home":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                chrome.tabs.sendMessage(layoutTabId, { type: "tilePage", direction: direction, url: (homePage != "") ? homePage : "about:blank" },{ frameId: 0 });
                
                break;
                
            case "tilepage-above-dup":
            case "tilepage-left-dup":
            case "tilepage-right-dup":
            case "tilepage-below-dup":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                chrome.tabs.sendMessage(layoutTabId, { type: "tileDupPage", direction: direction, url: info.pageUrl },{ frameId: 0 });
                
                break;
                
            case "tilepage-above-tiled-":
            case "tilepage-left-tiled-":
            case "tilepage-right-tiled-":
            case "tilepage-below-tiled-":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                chrome.tabs.sendMessage(layoutTabId, { type: "tilePage", direction: direction, url: tiledPageUrls[+matches[2]] },{ frameId: 0 });
                
                break;
                
            case "tilepage-above-tab-":
            case "tilepage-left-tab-":
            case "tilepage-right-tab-":
            case "tilepage-below-tab-":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                chrome.tabs.get(+matches[2],
                function(srctab)
                {
                    chrome.tabs.sendMessage(layoutTabId, { type: "tilePage", direction: direction, url: srctab.url },{ frameId: 0 });
                });
                
                break;
                
            case "assignpage-tiled-":
                
                chrome.tabs.sendMessage(layoutTabId, { type: "assignPage", url: tiledPageUrls[+matches[2]] },{ frameId: 0 });
                
                break;
                
            case "assignpage-tab-":
                
                chrome.tabs.get(+matches[2],
                function(srctab)
                {
                    chrome.tabs.sendMessage(layoutTabId, { type: "assignPage", url: srctab.url },{ frameId: 0 });
                });
                
                break;
                
            case "untilepage":
                
                chrome.tabs.sendMessage(layoutTabId, { type: "untilePage" },{ frameId: 0 });
                
                break;
                
            case "syncscroll":
                
                toggleSync(layoutTabId);

                break;
                
            case "inspecttile-this":
            case "inspecttile-before":
            case "inspecttile-after":
            case "inspecttile-parent":
            case "inspecttile-gparent":
            case "inspecttile-ggparent":
                
                chrome.tabs.sendMessage(layoutTabId, { type: "inspectTile", node: info.menuItemId.substr(12) },{ frameId: 0 });
                
                break;
            
            /* Tab context menu */
            
            case "tabmenu:tilepage-above":
            case "tabmenu:tilepage-left":
            case "tabmenu:tilepage-right":
            case "tabmenu:tilepage-below":
                
                direction = info.menuItemId.match(/above|left|right|below/)[0];
                
                url = (tab.id == layoutTabId) ? ((homePage != "") ? homePage : "about:blank") : tab.url;
                
                chrome.tabs.sendMessage(layoutTabId, { type: "tilePage", direction: direction, url: url },{ frameId: 0 });
                
                break;
                
            case "tabmenu:assignpage":
                
                url = (tab.id == layoutTabId) ? ((homePage != "") ? homePage : "about:blank") : tab.url;
                
                chrome.tabs.sendMessage(layoutTabId, { type: "assignPage", url: url },{ frameId: 0 });
                
                break;
        }
    });
    
    /* Tab event listeners */
    
    chrome.tabs.onActivated.addListener(  /* tab selected */
    function(activeInfo)
    {
        chrome.storage.local.get(null,
        function(object)
        {
            createContextMenu(activeInfo.tabId);
        });
    });
    
    chrome.tabs.onRemoved.addListener(  /* tab closed */
    function(tabId,removeInfo)
    {
        if (tabId == layoutTabId)
        {
            layoutTabId = -1;
            layoutTabReady = false;
            
            panelCount = 0;
            
            tiledPageUrls = new Array();
            tiledPageTitles = new Array();
            
            viewTiled = true;
            syncScroll = false;
            
            createContextMenu(tabId);
        }
    });
    
    /* Window event listeners */
    
    chrome.windows.onFocusChanged.addListener(  /* browser window gained focus or all browser windows lost focus */
    function(windowId)
    {
        if (windowId == chrome.windows.WINDOW_ID_NONE) return;  /* all Chrome windows have lost focus */
        
        // if (dialogWindowId != -1) chrome.windows.update(dialogWindowId,{ focused: true });  /* keep dialog window on top */
        
        chrome.tabs.query({ lastFocusedWindow: true, active: true },
        function(tabs)
        {
            createContextMenu(tabs[0].id);
        });
    });
    
    chrome.windows.onRemoved.addListener(  /* browser window closed */
    function(windowId)
    {
        if (windowId == dialogParentWindowId)  /* dialog parent window */
        {
            dialogParentWindowId = -1;
            
            chrome.windows.remove(dialogWindowId);
        }
        else if (windowId == dialogWindowId)  /* dialog window */
        {
            dialogWindowId = -1;
            dialogParentWindowId = -1;
        }
    });
    
    /* Web request listeners */
    
    chrome.webRequest.onBeforeRequest.addListener(
    function(details)
    {
        if (details.tabId == layoutTabId && details.type == "main_frame")
        {
            // return { cancel: "true" };  /*???*/
        }
    }
    ,{ urls: ["https://*/*"] },["blocking"]);
    
    chrome.webRequest.onBeforeSendHeaders.addListener(
    function(details)
    {
        if (details.tabId == layoutTabId && details.type == "main_frame")
        {
            // return { cancel: true };  /*???*/
        }
    }
    ,{ urls: ["<all_urls>"] },["blocking"]);
    
    chrome.webRequest.onHeadersReceived.addListener(
    function (details)
    {
        var i;
        
        if (details.tabId == layoutTabId && details.type == "sub_frame" && details.parentFrameId == 0)  /* parent frame is top-level frame in tab */
        {
            console.log("sub_frame  -  frameId: " + details.frameId + "  parentFrameId: " + details.parentFrameId + "  tabId: " + details.tabId);
            for (i = details.responseHeaders.length-1; i >= 0; i--)
            {
                if (xFrameOptions && details.responseHeaders[i].name.toLowerCase() == "x-frame-options")
                {
                    details.responseHeaders.splice(i,1);
                }
                else if (cspFrameAncestors && details.responseHeaders[i].name.toLowerCase() == "content-security-policy")
                {
                    details.responseHeaders[i].value = details.responseHeaders[i].value.replace(/frame-ancestors[^;]*(;|$)/g,"");
                }
            }
            
            return { responseHeaders: details.responseHeaders };
        }
    }
    , { urls: ["<all_urls>"] }, ["blocking", "responseHeaders"]);
    
    /* Message received listener */
    
    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        var i,tile;
        
        switch (message.type)
        {
            /* Messages from script in container page */
            
            case "layoutReady":
                
                layoutTabReady = true;
                
                break;
                
            case "panelCount":
                
                panelCount = message.count;
                
                chrome.browserAction.setBadgeText({ tabId: layoutTabId, text: "" + panelCount });            
                
                chrome.browserAction.setBadgeBackgroundColor({ tabId: layoutTabId, color: "#5070C0" });
                
                break;
                
            case "closeLayout":
                
                closeLayout();
                
                break;
                
            case "createBookmark":
                
                chrome.bookmarks.create({ url: message.url, title: message.title });
                
                break;
                
            case "openPageInTab":
                
                chrome.windows.getLastFocused({ },
                function(focuswin)
                {
                    chrome.tabs.create({ url: message.url, windowId: focuswin.id, active: true });
                });
                
                break;
                
            case "syncScroll":
                
                syncScroll = message.state;
                
                chrome.contextMenus.update("syncscroll",{ checked: syncScroll });
                
                break;
                
            case "viewTiled":
                
                viewTiled = message.state;
                
                break;
                
            case "updateContextMenu":
                
                createContextMenu(layoutTabId);
                
                break;
                
            case "updatePageDetails":
                
                tiledPageUrls = message.pageurls.slice(0);
                tiledPageTitles = message.pagetitles.slice(0);
                
                break;
                
            /* Messages from scripts in dialog pages */
            
            case "openLayoutReply":
                
                chrome.windows.remove(dialogWindowId,
                function()
                {
                    dialogWindowId = -1;
                    dialogParentWindowId = -1;
                    
                    if (message.result)
                    {
                        chrome.storage.local.get(null,
                        function(object)
                        {
                            var layoutString;
                            
                            layoutString = object["layout-string-" + message.layoutname];
                            
                            openLayout(message.layoutname,layoutString);
                        });
                    }
                });
                
                break;
                
            case "saveLayoutReply":
                
                chrome.windows.remove(dialogWindowId,
                function()
                {
                    dialogWindowId = -1;
                    dialogParentWindowId = -1;
                    
                    if (message.result)
                    {
                        chrome.tabs.sendMessage(layoutTabId,{ type: "saveLayout", layoutname: message.layoutname, layouttype: message.layouttype },{ frameId: 0 });
                    }
                });
                
                break;
                
            case "deleteLayoutReply":
                
                chrome.windows.remove(dialogWindowId,
                function()
                {
                    dialogWindowId = -1;
                    dialogParentWindowId = -1;
                    
                    if (message.result)
                    {
                        chrome.storage.local.remove("layout-string-" + message.layoutname);
                    }
                });
                
                break;
                
            case "defaultLayoutReply":
                
                chrome.windows.remove(dialogWindowId,
                function()
                {
                    dialogWindowId = -1;
                    dialogParentWindowId = -1;
                    
                    if (message.result)
                    {
                        clearDefaultLayoutCheckMarks();
                        
                        chrome.contextMenus.update("defaultlayout-saved",{ checked: true });
                        
                        chrome.storage.local.set({ "layout-defaultdesign" : "saved", "layout-defaultname": message.layoutname });
                    }
                });
                
                break;
        }
    });
}

/************************************************************************/

/* New Layout */

function newLayout(design)
{
    chrome.tabs.query({ },
    function(tabs)
    {
        var panelCount;
        var pageUrls = new Array();
        
        if (design.substr(0,1) == "2") panelCount = 2;
        else if (design.substr(0,1) == "3") panelCount = 3;
        else if (design.substr(0,1) == "4") panelCount = 4;
        else if (design.substr(0,1) == "6") panelCount = 6;
        else if (design.substr(0,3) == "all") panelCount = -1;
        
        pageUrls = enumeratePageUrls(tabs,panelCount);
        
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            chrome.tabs.create({ url: chrome.runtime.getURL("container.html"), windowId: focuswin.id },
            function(tab)
            {
                var timer;
                
                layoutTabId = tab.id;
            
                timer = window.setInterval(
                function()
                {
                    if (layoutTabReady)
                    {
                        chrome.tabs.sendMessage(layoutTabId, { type: "newLayout", layout: design, pageurls: pageUrls, tabid: tab.id, incognito: tab.incognito },{ frameId: 0 });
                        
                        window.clearInterval(timer);
                    }
                },50);
            });
        });
    });
}

/************************************************************************/

/* Close Layout */

function closeLayout()
{
    chrome.tabs.remove(layoutTabId);
    
    /* chrome.tabs.remove() may be prevented by container's onbeforeunload listener */
    /* so remainder of tidying up is done in chrome.tabs.onRemoved.addListener() */
}

/************************************************************************/

/* Toggle Sync */

function toggleSync(tabId)
{
    chrome.tabs.sendMessage(tabId, { type: "toggleSync" },{ frameId: 0 });
}

/************************************************************************/

/* Toggle View */

function toggleView(tabId)
{
    chrome.tabs.sendMessage(tabId, { type: "toggleView" },{ frameId: 0 });
}

/************************************************************************/

/* Open Layout */

function openLayoutDialog()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        dialogParentWindowLeft = focuswin.left;
        dialogParentWindowTop = focuswin.top;
        dialogParentWindowHeight = focuswin.height;
        dialogParentWindowWidth = focuswin.width;
        
        width = 330 + lrbMargin*2;
        height = 338 + topMargin + lrbMargin + ((platformOS != "win") ? -28 : 0);  /* less header height if not Windows */
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.runtime.getURL("openlayout.html"), /* focused: null, */  /* Firefox - focused not supported */  /*???*/
                                left: left, top: top,  /* Firefox - left and top not supported - window opens in middle of screen */  /*???*/
                                width: width, height: height },
        function(newwin)
        {
            dialogWindowId = newwin.id;
        });
    });
}

function openLayout(layoutName,layoutString)
{
    chrome.tabs.query({ },
    function(tabs)
    {
        var i,panelCount;
        var pageUrls = new Array();
        
        panelCount = +layoutString.match(/\d+,(\d+);/)[1];
        
        pageUrls = enumeratePageUrls(tabs,panelCount);
        
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            chrome.tabs.create({ url: chrome.runtime.getURL("container.html"), windowId: focuswin.id },
            function(tab)
            {
                var timer;
                
                layoutTabId = tab.id;
                
                timer = window.setInterval(
                function()
                {
                    if (layoutTabReady)
                    {
                        chrome.tabs.sendMessage(layoutTabId,{ type: "openLayout", layoutstring: layoutString, layoutname: layoutName, pageurls: pageUrls, tabid: tab.id, incognito: tab.incognito },{ frameId: 0 });
                        
                        window.clearInterval(timer);
                    }
                },50);
            });
        });
    });
}

/************************************************************************/

/* Save Layout */

function saveLayoutDialog()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        dialogParentWindowLeft = focuswin.left;
        dialogParentWindowTop = focuswin.top;
        dialogParentWindowHeight = focuswin.height;
        dialogParentWindowWidth = focuswin.width;
        
        width = 330 + lrbMargin*2;
        height = 408 + topMargin + lrbMargin + ((platformOS != "win") ? -26 : 0);  /* less header height if not Windows */
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.runtime.getURL("savelayout.html"), /* focused: null, */  /* Firefox - focused not supported */  /*???*/
                                left: left, top: top,  /* Firefox - left and top not supported - window opens in middle of screen */  /*???*/
                                width: width, height: height },
        function(newwin)
        {
            dialogWindowId = newwin.id;
        });
    });
}

/************************************************************************/

/* Delete Layout */

function deleteLayoutDialog()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        dialogParentWindowLeft = focuswin.left;
        dialogParentWindowTop = focuswin.top;
        dialogParentWindowHeight = focuswin.height;
        dialogParentWindowWidth = focuswin.width;
        
        width = 330 + lrbMargin*2;
        height = 338 + topMargin + lrbMargin + ((platformOS != "win") ? -28 : 0);  /* less header height if not Windows */
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.runtime.getURL("deletelayout.html"), /* focused: null, */  /* Firefox - focused not supported */  /*???*/
                                left: left, top: top,  /* Firefox - left and top not supported - window opens in middle of screen */  /*???*/
                                width: width, height: height },
        function(newwin)
        {
            dialogWindowId = newwin.id;
        });
    });
}

/************************************************************************/

/* Default Layout */

function defaultLayoutDialog()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        dialogParentWindowLeft = focuswin.left;
        dialogParentWindowTop = focuswin.top;
        dialogParentWindowHeight = focuswin.height;
        dialogParentWindowWidth = focuswin.width;
        
        width = 330 + lrbMargin*2;
        height = 338 + topMargin + lrbMargin + ((platformOS != "win") ? -28 : 0);  /* less header height if not Windows */
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.runtime.getURL("defaultlayout.html"), /* focused: null, */  /* Firefox - focused not supported */  /*???*/
                                left: left, top: top,  /* Firefox - left and top not supported - window opens in middle of screen */  /*???*/
                                width: width, height: height },
        function(newwin)
        {
            dialogWindowId = newwin.id;
        });
    });
}

/************************************************************************/

/* Create context menu */

function createContextMenu(tabId)
{
    chrome.tabs.query({ },
    function(tabs)
    {
        var i,context;
        
        chrome.contextMenus.removeAll();
        
        /* Page or button context menu */
        
        if (tabId == layoutTabId)  /* layout tab */
        {
            context = showSubmenu ? "all" : "browser_action";
            
            chrome.contextMenus.create({ id: "newcloselayout", title: "Close Layout", contexts: [ context ], enabled: true });
            chrome.contextMenus.create({ id: "managelayouts", title: "Manage Layouts", contexts: [ context ], enabled: true });
            chrome.contextMenus.create({ id: "tilepage", title: "Tile Page", contexts: [ context ], enabled: viewTiled });
            chrome.contextMenus.create({ id: "assignpage", title: "Assign Page", contexts: [ context ], enabled: viewTiled });
            chrome.contextMenus.create({ id: "untilepage", title: "Untile Page", contexts: [ context ], enabled: viewTiled });
            chrome.contextMenus.create({ id: "syncscroll", title: "Sync Scroll", type: "checkbox", contexts: [ context ], enabled: viewTiled, checked: syncScroll });
            
            if (enableInspect) chrome.contextMenus.create({ id: "inspecttile", title: "Inspect Tile", contexts: [ context ], enabled: viewTiled });
            
            createManageLayoutsSubmenu(tabId);
            
            chrome.contextMenus.create({ id: "tilepage-above", parentId: "tilepage", title: "Above", contexts: [ "all" ], enabled: true });
            chrome.contextMenus.create({ id: "tilepage-left", parentId: "tilepage", title: "Left", contexts: [ "all" ], enabled: true });
            chrome.contextMenus.create({ id: "tilepage-right", parentId: "tilepage", title: "Right", contexts: [ "all" ], enabled: true });
            chrome.contextMenus.create({ id: "tilepage-below", parentId: "tilepage", title: "Below", contexts: [ "all" ], enabled: true });
            
            createTilePageSubmenu(tabs,"above");
            createTilePageSubmenu(tabs,"left");
            createTilePageSubmenu(tabs,"right");
            createTilePageSubmenu(tabs,"below");
            
            createAssignPageSubmenu(tabs,"above");
            
            if (enableInspect) createInspectTileSubmenu();
        }
        else
        {
            context = showSubmenu ? "all" : "browser_action";
            
            if (layoutTabId == -1) chrome.contextMenus.create({ id: "newcloselayout", title: "New Layout", contexts: [ context ], enabled: true });
            else chrome.contextMenus.create({ id: "newcloselayout", title: "Close Layout", contexts: [ context ], enabled: true });
            chrome.contextMenus.create({ id: "managelayouts", title: "Manage Layouts", contexts: [ context ], enabled: true });
            chrome.contextMenus.create({ id: "tilepage", title: "Tile Page", contexts: [ context ], enabled: false });
            chrome.contextMenus.create({ id: "assignpage", title: "Assign Page", contexts: [ context ], enabled: false });
            chrome.contextMenus.create({ id: "untilepage", title: "Untile Page", contexts: [ context ], enabled: false });
            chrome.contextMenus.create({ id: "syncscroll", title: "Sync Scroll", type: "checkbox", contexts: [ context ], enabled: false, checked: syncScroll });
            
            if (layoutTabId == -1) createNewLayoutSubmenu();
            
            createManageLayoutsSubmenu(tabId);
        }
        
        /* Tab context menu */
        
        if (tabId == layoutTabId)  /* layout tab */
        {
            if (isFirefox && showSubmenu)
            {
                chrome.contextMenus.create({ id: "tabmenu:tilepage", title: "Tile Page", contexts: [ "tab" ], enabled: viewTiled });
                chrome.contextMenus.create({ id: "tabmenu:assignpage", title: "Assign Page", contexts: [ "tab" ], enabled: viewTiled });
                
                chrome.contextMenus.create({ id: "tabmenu:tilepage-above", parentId: "tabmenu:tilepage", title: "Above", contexts: [ "tab" ], enabled: true });
                chrome.contextMenus.create({ id: "tabmenu:tilepage-left", parentId: "tabmenu:tilepage", title: "Left", contexts: [ "tab" ], enabled: true });
                chrome.contextMenus.create({ id: "tabmenu:tilepage-right", parentId: "tabmenu:tilepage", title: "Right", contexts: [ "tab" ], enabled: true });
                chrome.contextMenus.create({ id: "tabmenu:tilepage-below", parentId: "tabmenu:tilepage", title: "Below", contexts: [ "tab" ], enabled: true });
            }
        }
    });
}

function createNewLayoutSubmenu()
{
    chrome.contextMenus.create({ id: "newlayout-default", parentId: "newcloselayout", title: "Default Layout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-sep-2", type: "separator", parentId: "newcloselayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-2vert", parentId: "newcloselayout", title: "2 Pages - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-2horiz", parentId: "newcloselayout", title: "2 Pages - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-sep-3", type: "separator", parentId: "newcloselayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-3vert", parentId: "newcloselayout", title: "3 Pages - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-3horiz", parentId: "newcloselayout", title: "3 Pages - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-sep-4", type: "separator", parentId: "newcloselayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-4vertgrid", parentId: "newcloselayout", title: "4 Pages - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-4horizgrid", parentId: "newcloselayout", title: "4 Pages - Horizontal Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-sep-6", type: "separator", parentId: "newcloselayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-6vertgrid", parentId: "newcloselayout", title: "6 Pages - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-6horizgrid", parentId: "newcloselayout", title: "6 Pages - Horizontal Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-sep-all", type: "separator", parentId: "newcloselayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-allvert", parentId: "newcloselayout", title: "All Pages - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-allhoriz", parentId: "newcloselayout", title: "All Pages - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-allvertgrid", parentId: "newcloselayout", title: "All Pages - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "newlayout-allhorizgrid", parentId: "newcloselayout", title: "All Pages - Horizontal Grid", contexts: [ "all" ], enabled: true });
}

function createManageLayoutsSubmenu(tabId)
{
    
    if (layoutTabId == -1) chrome.contextMenus.create({ id: "opensavelayout", parentId: "managelayouts", title: "Open Layout...", contexts: [ "all" ], enabled: true });
    else if (tabId == layoutTabId) chrome.contextMenus.create({ id: "opensavelayout", parentId: "managelayouts", title: "Save Layout...", contexts: [ "all" ], enabled: true });
    else  chrome.contextMenus.create({ id: "opensavelayout", parentId: "managelayouts", title: "Open Layout...", contexts: [ "all" ], enabled: false });
    
    chrome.contextMenus.create({ id: "deletelayout", parentId: "managelayouts", title: "Delete Layout...", contexts: [ "all" ], enabled: true });
    
    chrome.contextMenus.create({ id: "defaultlayout", parentId: "managelayouts", title: "Default Layout...", contexts: [ "all" ], enabled: true });
    
    chrome.contextMenus.create({ id: "defaultlayout-saved", type: "checkbox", parentId: "defaultlayout", title: "Saved Layout...", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-sep-2", type: "separator", parentId: "defaultlayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-2vert", type: "checkbox", parentId: "defaultlayout", title: "2 Tabs - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-2horiz", type: "checkbox", parentId: "defaultlayout", title: "2 Tabs - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-sep-3", type: "separator", parentId: "defaultlayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-3vert", type: "checkbox", parentId: "defaultlayout", title: "3 Tabs - Vertical", contexts: [ "all" ], enabled: true, checked: false });
    chrome.contextMenus.create({ id: "defaultlayout-3horiz", type: "checkbox", parentId: "defaultlayout", title: "3 Tabs - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-sep-4", type: "separator", parentId: "defaultlayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-4vertgrid", type: "checkbox", parentId: "defaultlayout", title: "4 Tabs - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-4horizgrid", type: "checkbox", parentId: "defaultlayout", title: "4 Tabs - Horizontal Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-sep-6", type: "separator", parentId: "defaultlayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-6vertgrid", type: "checkbox", parentId: "defaultlayout", title: "6 Tabs - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-6horizgrid", type: "checkbox", parentId: "defaultlayout", title: "6 Tabs - Horizontal Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-sep-all", type: "separator", parentId: "defaultlayout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-allvert", type: "checkbox", parentId: "defaultlayout", title: "All Tabs - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-allhoriz", type: "checkbox", parentId: "defaultlayout", title: "All Tabs - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-allvertgrid", type: "checkbox", parentId: "defaultlayout", title: "All Tabs - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "defaultlayout-allhorizgrid", type: "checkbox", parentId: "defaultlayout", title: "All Tabs - Horizontal Grid", contexts: [ "all" ], enabled: true });
    
    window.setTimeout(
    function()
    {
        chrome.storage.local.get(null,
        function(object)
        {
            chrome.contextMenus.update("defaultlayout-" + object["layout-defaultdesign"],{ checked: true });
        });
    },10);
}

function createTilePageSubmenu(tabs,dir)
{
    var i;
    
    /* Special pages */
    
    chrome.contextMenus.create({ id: "tilepage-" + dir + "-link", parentId: "tilepage-" + dir, title: "This Link", contexts: [ "link" ], enabled: true });
    chrome.contextMenus.create({ id: "tilepage-" + dir + "-home", parentId: "tilepage-" + dir, title: "Home Page", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "tilepage-" + dir + "-dup", parentId: "tilepage-" + dir, title: "Duplicate Page", contexts: [ "all" ], enabled: true });
    
    /* Separator */
    
    chrome.contextMenus.create({ id: "tilepage-" + dir + "-sep-1", type: "separator", parentId: "tilepage-" + dir, contexts: [ "all" ], enabled: true });
    
    /* Tile pages */
    
    for (i = 0; i < tiledPageTitles.length; i++)
    {
        chrome.contextMenus.create({ id: "tilepage-" + dir + "-tiled-" + i, parentId: "tilepage-" + dir, title: tiledPageTitles[i], contexts: [ "all" ], enabled: true });
    }
    
    /* Separator */
    
    chrome.contextMenus.create({ id: "tilepage-" + dir + "-sep-2", type: "separator", parentId: "tilepage-" + dir, contexts: [ "all" ], enabled: true });
    
    /* Tab pages */
    
    for (i = 0; i < tabs.length; i++)
    {
        if (tabs[i].id != layoutTabId)
        {
            chrome.contextMenus.create({ id: "tilepage-" + dir + "-tab-" + tabs[i].id, parentId: "tilepage-" + dir, title: tabs[i].title, contexts: [ "all" ], enabled: true });
        }
    }
}
    
function createAssignPageSubmenu(tabs,dir)
{
    var i;
    
    /* Tile pages */
    
    for (i = 0; i < tiledPageTitles.length; i++)
    {
        chrome.contextMenus.create({ id: "assignpage-tiled-" + i, parentId: "assignpage", title: tiledPageTitles[i], contexts: [ "all" ], enabled: true });
    }
    
    /* Separator */
    
    chrome.contextMenus.create({ id: "assignpage-sep-1", type: "separator", parentId: "assignpage", contexts: [ "all" ], enabled: true });
    
    /* Tab pages */
    
    for (i = 0; i < tabs.length; i++)
    {
        if (tabs[i].id != layoutTabId)
        {
            chrome.contextMenus.create({ id: "assignpage-tab-" + tabs[i].id, parentId: "assignpage", title: tabs[i].title, contexts: [ "all" ], enabled: true });
        }
    }
}

function createInspectTileSubmenu()
{
    chrome.contextMenus.create({ id: "inspecttile-this", parentId: "inspecttile", title: "This", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "inspecttile-before", parentId: "inspecttile", title: "Before", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "inspecttile-after", parentId: "inspecttile", title: "After", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "inspecttile-parent", parentId: "inspecttile", title: "Parent", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "inspecttile-gparent", parentId: "inspecttile", title: "G-Parent", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "inspecttile-ggparent", parentId: "inspecttile", title: "G-G-Parent", contexts: [ "all" ], enabled: true });
}

function clearDefaultLayoutCheckMarks()
{
    chrome.contextMenus.update("defaultlayout-saved",{ checked: false });
    chrome.contextMenus.update("defaultlayout-2vert",{ checked: false });
    chrome.contextMenus.update("defaultlayout-2horiz",{ checked: false });
    chrome.contextMenus.update("defaultlayout-3vert",{ checked: false });
    chrome.contextMenus.update("defaultlayout-3horiz",{ checked: false });
    chrome.contextMenus.update("defaultlayout-4vertgrid",{ checked: false });
    chrome.contextMenus.update("defaultlayout-4horizgrid",{ checked: false });
    chrome.contextMenus.update("defaultlayout-6vertgrid",{ checked: false });
    chrome.contextMenus.update("defaultlayout-6horizgrid",{ checked: false });
    chrome.contextMenus.update("defaultlayout-allvert",{ checked: false });
    chrome.contextMenus.update("defaultlayout-allhoriz",{ checked: false });
    chrome.contextMenus.update("defaultlayout-allvertgrid",{ checked: false });
    chrome.contextMenus.update("defaultlayout-allhorizgrid",{ checked: false });
}

/************************************************************************/

/* Utility functions */

function enumeratePageUrls(tabs,panelCount)
{
    var i,multiple,pos,selpos,startpos,endpos;
    var validTabs = new Array();
    var pageUrls = new Array();
    
    pos = 0;
    multiple = false;
    
    for (i = 0; i < tabs.length; i++)
    {
        if (!restrictedPage(tabs[i].url))
        {
            validTabs[pos] = { url: tabs[i].url, highlighted: tabs[i].highlighted };
        
            if (tabs[i].active) selpos = pos;  /* selected tab */
            else if (tabs[i].highlighted) multiple = true;  /* highlighted tab */
            
            pos++;
        }
    }
    
    if (panelCount == -1)  /* New Layout - assign all tab pages */
    {
        for (pos = 0; pos < validTabs.length; pos++)
        {
            pageUrls.push(validTabs[pos].url);
        }
    }
    else if (multiple)  /* New Layout - assign selected tab page and highlighted tab pages */
    {
        for (pos = 0; pos < validTabs.length; pos++)
        {
            if (validTabs[pos].highlighted && ((pos < selpos && pageUrls.length < panelCount-1) || (pos >= selpos && pageUrls.length < panelCount)))
            {
                pageUrls.push(validTabs[pos].url);
            }
        }
    }
    else if (assignTabs)  /* New/Open Layout - assign selected tab page and tab pages from left or right of selected tab */
    {
        if (assignLeft)
        {
            if (selpos+1 >= panelCount) { startpos = selpos-panelCount+1; endpos = selpos; }
            else if (validTabs.length >= panelCount) { startpos = 0; endpos = panelCount-1; }
            else { startpos = 0; endpos = validTabs.length-1; }
            
            for (pos = startpos; pos <= endpos; pos++)
            {
                pageUrls.push(validTabs[pos].url);
            }
        }
        else
        {
            if (validTabs.length-selpos >= panelCount) { startpos = selpos; endpos = selpos+panelCount-1; }
            else if (validTabs.length >= panelCount) { startpos = validTabs.length-panelCount; endpos = validTabs.length-1; }
            else { startpos = 0; endpos = validTabs.length-1; }
            
            for (pos = startpos; pos <= endpos; pos++)
            {
                pageUrls.push(validTabs[pos].url);
            }
        }
    }
    
    return pageUrls;
}

function restrictedPage(url)
{
    /* Pages that cannot be loaded into an <iframe> (1) */
    /* Pages that do not allow content scripts to be loaded (2) */
    /* Pages that are not allowed by WebExtensions Content Security Policy (3) */
    
    return (url.substr(0,6) == "about:" || url.substr(0,7) == "chrome:" ||  /*(1)*/
            url.substr(0,5) == "data:" || url.substr(0,11) == "javascript:" ||  /*(2)*/
            url.substr(0,14) == "moz-extension:" || url.substr(0,26) == "https://addons.mozilla.org" ||  /*(3)*/
            url.substr(0,17) == "chrome-extension:" || url.substr(0,34) == "https://chrome.google.com/webstore");  /*(3)*/
}

function webRequestLog(tag,details)
{
    var time,tabid,mode,type,url;
    
    time = ("        " + performance.now().toFixed(2)).substr(-8);
    tabid = ("  " + details.tabId).substr(-2);
    type = (details.type + "          ").substr(0,10);
    url = details.url.substr(0,50);
    
    console.log(tag + ":  " + time + "  " + tabid + "   " + type + "   " + url);
}

/************************************************************************/

/* Display debug notification */

function debugNotify(message)
{
    chrome.notifications.create("debug",{ type: "basic", iconUrl: "icon32.png", title: "TILE TABS WE - DEBUG", message: "" + message });
}

function debugModal(message)
{
    chrome.tabs.query({ lastFocusedWindow: true, active: true },
    function(tabs)
    {
        chrome.tabs.executeScript(tabs[0].id,{ code: "alert('" + message + "');" });
    });
}

/************************************************************************/
